@@ -9,6 +9,7 @@ package db
 import (
 	"bytes"
 	"encoding/binary"
+	"fmt"
 	"os"
 	"sort"
 	"strings"
@@ -35,10 +36,13 @@ type Instance struct {
 }
 
 const (
-	keyPrefixLen = 1
-	keyFolderLen = 4 // indexed
-	keyDeviceLen = 4 // indexed
-	keyHashLen   = 32
+	keyPrefixLen   = 1
+	keyFolderLen   = 4 // indexed
+	keyDeviceLen   = 4 // indexed
+	keySequenceLen = 8
+	keyHashLen     = 32
+
+	maxInt64 int64 = 1<<63 - 1
 )
 
 func Open(file string) (*Instance, error) {
@@ -57,44 +61,32 @@ func Open(file string) (*Instance, error) {
 		// the database and reindexing...
 		l.Infoln("Database corruption detected, unable to recover. Reinitializing...")
 		if err := os.RemoveAll(file); err != nil {
-			return nil, err
+			return nil, errorSuggestion{err, "failed to delete corrupted database"}
 		}
 		db, err = leveldb.OpenFile(file, opts)
 	}
 	if err != nil {
-		return nil, err
+		return nil, errorSuggestion{err, "is another instance of Syncthing running?"}
 	}
 
-	return newDBInstance(db, file), nil
+	return newDBInstance(db, file)
 }
 
 func OpenMemory() *Instance {
 	db, _ := leveldb.Open(storage.NewMemStorage(), nil)
-	return newDBInstance(db, "<memory>")
+	ldb, _ := newDBInstance(db, "<memory>")
+	return ldb
 }
 
-func newDBInstance(db *leveldb.DB, location string) *Instance {
+func newDBInstance(db *leveldb.DB, location string) (*Instance, error) {
 	i := &Instance{
 		DB:       db,
 		location: location,
 	}
 	i.folderIdx = newSmallIndex(i, []byte{KeyTypeFolderIdx})
 	i.deviceIdx = newSmallIndex(i, []byte{KeyTypeDeviceIdx})
-	return i
-}
-
-// UpdateSchema does transitions to the current db version if necessary
-func (db *Instance) UpdateSchema() {
-	miscDB := NewNamespacedKV(db, string(KeyTypeMiscData))
-	prevVersion, _ := miscDB.Int64("dbVersion")
-
-	if prevVersion == 0 {
-		db.updateSchema0to1()
-	}
-
-	if prevVersion != dbVersion {
-		miscDB.PutInt64("dbVersion", dbVersion)
-	}
+	err := i.updateSchema()
+	return i, err
 }
 
 // Committed returns the number of items committed to the database since startup
@@ -112,9 +104,10 @@ func (db *Instance) updateFiles(folder, device []byte, fs []protocol.FileInfo, m
 	defer t.close()
 
 	var fk []byte
+	var gk []byte
 	for _, f := range fs {
 		name := []byte(f.Name)
-		fk = db.deviceKeyInto(fk[:cap(fk)], folder, device, name)
+		fk = db.deviceKeyInto(fk, folder, device, name)
 
 		// Get and unmarshal the file entry. If it doesn't exist or can't be
 		// unmarshalled we'll add it as a new entry.
@@ -124,8 +117,9 @@ func (db *Instance) updateFiles(folder, device []byte, fs []protocol.FileInfo, m
 			err = ef.Unmarshal(bs)
 		}
 
-		// The Invalid flag might change without the version being bumped.
-		if err == nil && ef.Version.Equal(f.Version) && ef.Invalid == f.Invalid {
+		// Local flags or the invalid bit might change without the version
+		// being bumped. The IsInvalid() method handles both.
+		if err == nil && ef.Version.Equal(f.Version) && ef.IsInvalid() == f.IsInvalid() {
 			continue
 		}
 
@@ -135,8 +129,10 @@ func (db *Instance) updateFiles(folder, device []byte, fs []protocol.FileInfo, m
 		}
 		meta.addFile(devID, f)
 
-		t.insertFile(folder, device, f)
-		t.updateGlobal(folder, device, f, meta)
+		t.insertFile(fk, folder, device, f)
+
+		gk = db.globalKeyInto(gk, folder, name)
+		t.updateGlobal(gk, folder, device, f, meta)
 
 		// Write out and reuse the batch every few records, to avoid the batch
 		// growing too large and thus allocating unnecessarily much memory.
@@ -144,21 +140,56 @@ func (db *Instance) updateFiles(folder, device []byte, fs []protocol.FileInfo, m
 	}
 }
 
+func (db *Instance) addSequences(folder []byte, fs []protocol.FileInfo) {
+	t := db.newReadWriteTransaction()
+	defer t.close()
+
+	var sk []byte
+	var dk []byte
+	for _, f := range fs {
+		sk = db.sequenceKeyInto(sk, folder, f.Sequence)
+		dk = db.deviceKeyInto(dk, folder, protocol.LocalDeviceID[:], []byte(f.Name))
+		t.Put(sk, dk)
+		l.Debugf("adding sequence; folder=%q sequence=%v %v", folder, f.Sequence, f.Name)
+		t.checkFlush()
+	}
+}
+
+func (db *Instance) removeSequences(folder []byte, fs []protocol.FileInfo) {
+	t := db.newReadWriteTransaction()
+	defer t.close()
+
+	var sk []byte
+	for _, f := range fs {
+		t.Delete(db.sequenceKeyInto(sk, folder, f.Sequence))
+		l.Debugf("removing sequence; folder=%q sequence=%v %v", folder, f.Sequence, f.Name)
+		t.checkFlush()
+	}
+}
+
 func (db *Instance) withHave(folder, device, prefix []byte, truncate bool, fn Iterator) {
+	if len(prefix) > 0 {
+		unslashedPrefix := prefix
+		if bytes.HasSuffix(prefix, []byte{'/'}) {
+			unslashedPrefix = unslashedPrefix[:len(unslashedPrefix)-1]
+		} else {
+			prefix = append(prefix, '/')
+		}
+
+		if f, ok := db.getFileTrunc(db.deviceKey(folder, device, unslashedPrefix), true); ok && !fn(f) {
+			return
+		}
+	}
+
 	t := db.newReadOnlyTransaction()
 	defer t.close()
 
 	dbi := t.NewIterator(util.BytesPrefix(db.deviceKey(folder, device, prefix)[:keyPrefixLen+keyFolderLen+keyDeviceLen+len(prefix)]), nil)
 	defer dbi.Release()
 
-	slashedPrefix := prefix
-	if !bytes.HasSuffix(prefix, []byte{'/'}) {
-		slashedPrefix = append(slashedPrefix, '/')
-	}
-
 	for dbi.Next() {
 		name := db.deviceKeyName(dbi.Key())
-		if len(prefix) > 0 && !bytes.Equal(name, prefix) && !bytes.HasPrefix(name, slashedPrefix) {
+		if len(prefix) > 0 && !bytes.HasPrefix(name, prefix) {
 			return
 		}
 
@@ -171,7 +202,34 @@ func (db *Instance) withHave(folder, device, prefix []byte, truncate bool, fn It
 			l.Debugln("unmarshal error:", err)
 			continue
 		}
-		if cont := fn(f); !cont {
+		if !fn(f) {
+			return
+		}
+	}
+}
+
+func (db *Instance) withHaveSequence(folder []byte, startSeq int64, fn Iterator) {
+	t := db.newReadOnlyTransaction()
+	defer t.close()
+
+	dbi := t.NewIterator(&util.Range{Start: db.sequenceKey(folder, startSeq), Limit: db.sequenceKey(folder, maxInt64)}, nil)
+	defer dbi.Release()
+
+	for dbi.Next() {
+		f, ok := db.getFile(dbi.Value())
+		if !ok {
+			l.Debugln("missing file for sequence number", db.sequenceKeySequence(dbi.Key()))
+			continue
+		}
+
+		if shouldDebug() {
+			key := dbi.Key()
+			seq := int64(binary.BigEndian.Uint64(key[keyPrefixLen+keyFolderLen:]))
+			if f.Sequence != seq {
+				panic(fmt.Sprintf("sequence index corruption, file sequence %d != expected %d", f.Sequence, seq))
+			}
+		}
+		if !fn(f) {
 			return
 		}
 	}
@@ -184,6 +242,8 @@ func (db *Instance) withAllFolderTruncated(folder []byte, fn func(device []byte,
 	dbi := t.NewIterator(util.BytesPrefix(db.deviceKey(folder, nil, nil)[:keyPrefixLen+keyFolderLen]), nil)
 	defer dbi.Release()
 
+	var gk []byte
+
 	for dbi.Next() {
 		device := db.deviceKeyDevice(dbi.Key())
 		var f FileInfoTruncated
@@ -200,106 +260,114 @@ func (db *Instance) withAllFolderTruncated(folder []byte, fn func(device []byte,
 		switch f.Name {
 		case "", ".", "..", "/": // A few obviously invalid filenames
 			l.Infof("Dropping invalid filename %q from database", f.Name)
-			t.removeFromGlobal(folder, device, nil, nil)
+			name := []byte(f.Name)
+			gk = db.globalKeyInto(gk, folder, name)
+			t.removeFromGlobal(gk, folder, device, name, nil)
 			t.Delete(dbi.Key())
 			t.checkFlush()
 			continue
 		}
 
-		if cont := fn(device, f); !cont {
+		if !fn(device, f) {
 			return
 		}
 	}
 }
 
-func (db *Instance) getFile(folder, device, file []byte) (protocol.FileInfo, bool) {
-	return getFile(db, db.deviceKey(folder, device, file))
+func (db *Instance) getFile(key []byte) (protocol.FileInfo, bool) {
+	if f, ok := db.getFileTrunc(key, false); ok {
+		return f.(protocol.FileInfo), true
+	}
+	return protocol.FileInfo{}, false
 }
 
-func (db *Instance) getGlobal(folder, file []byte, truncate bool) (FileIntf, bool) {
-	k := db.globalKey(folder, file)
-
-	t := db.newReadOnlyTransaction()
-	defer t.close()
-
-	bs, err := t.Get(k, nil)
-	if err != nil {
-		return nil, false
-	}
-
-	var vl VersionList
-	err = vl.Unmarshal(bs)
+func (db *Instance) getFileTrunc(key []byte, trunc bool) (FileIntf, bool) {
+	bs, err := db.Get(key, nil)
 	if err == leveldb.ErrNotFound {
 		return nil, false
 	}
 	if err != nil {
-		l.Debugln("unmarshal error:", k, err)
-		return nil, false
-	}
-	if len(vl.Versions) == 0 {
-		l.Debugln("no versions:", k)
+		l.Debugln("surprise error:", err)
 		return nil, false
 	}
 
-	k = db.deviceKey(folder, vl.Versions[0].Device, file)
-	bs, err = t.Get(k, nil)
+	f, err := unmarshalTrunc(bs, trunc)
 	if err != nil {
-		l.Debugln("surprise error:", k, err)
+		l.Debugln("unmarshal error:", err)
 		return nil, false
 	}
+	return f, true
+}
+
+func (db *Instance) getGlobal(folder, file []byte, truncate bool) (FileIntf, bool) {
+	t := db.newReadOnlyTransaction()
+	defer t.close()
+
+	_, _, f, ok := db.getGlobalInto(t, nil, nil, folder, file, truncate)
+	return f, ok
+}
+
+func (db *Instance) getGlobalInto(t readOnlyTransaction, gk, dk, folder, file []byte, truncate bool) ([]byte, []byte, FileIntf, bool) {
+	gk = db.globalKeyInto(gk, folder, file)
+
+	bs, err := t.Get(gk, nil)
+	if err != nil {
+		return gk, dk, nil, false
+	}
 
-	fi, err := unmarshalTrunc(bs, truncate)
-	if err != nil {
-		l.Debugln("unmarshal error:", k, err)
-		return nil, false
+	vl, ok := unmarshalVersionList(bs)
+	if !ok {
+		return gk, dk, nil, false
 	}
-	return fi, true
+
+	dk = db.deviceKeyInto(dk, folder, vl.Versions[0].Device, file)
+	if fi, ok := db.getFileTrunc(dk, truncate); ok {
+		return gk, dk, fi, true
+	}
+
+	return gk, dk, nil, false
 }
 
 func (db *Instance) withGlobal(folder, prefix []byte, truncate bool, fn Iterator) {
+	if len(prefix) > 0 {
+		unslashedPrefix := prefix
+		if bytes.HasSuffix(prefix, []byte{'/'}) {
+			unslashedPrefix = unslashedPrefix[:len(unslashedPrefix)-1]
+		} else {
+			prefix = append(prefix, '/')
+		}
+
+		if f, ok := db.getGlobal(folder, unslashedPrefix, truncate); ok && !fn(f) {
+			return
+		}
+	}
+
 	t := db.newReadOnlyTransaction()
 	defer t.close()
 
 	dbi := t.NewIterator(util.BytesPrefix(db.globalKey(folder, prefix)), nil)
 	defer dbi.Release()
 
-	slashedPrefix := prefix
-	if !bytes.HasSuffix(prefix, []byte{'/'}) {
-		slashedPrefix = append(slashedPrefix, '/')
-	}
-
 	var fk []byte
 	for dbi.Next() {
-		var vl VersionList
-		err := vl.Unmarshal(dbi.Value())
-		if err != nil {
-			l.Debugln("unmarshal error:", err)
-			continue
-		}
-		if len(vl.Versions) == 0 {
-			l.Debugln("no versions:", dbi.Key())
-			continue
-		}
-
 		name := db.globalKeyName(dbi.Key())
-		if len(prefix) > 0 && !bytes.Equal(name, prefix) && !bytes.HasPrefix(name, slashedPrefix) {
+		if len(prefix) > 0 && !bytes.HasPrefix(name, prefix) {
 			return
 		}
 
-		fk = db.deviceKeyInto(fk[:cap(fk)], folder, vl.Versions[0].Device, name)
-		bs, err := t.Get(fk, nil)
-		if err != nil {
-			l.Debugln("surprise error:", err)
+		vl, ok := unmarshalVersionList(dbi.Value())
+		if !ok {
 			continue
 		}
 
-		f, err := unmarshalTrunc(bs, truncate)
-		if err != nil {
-			l.Debugln("unmarshal error:", err)
+		fk = db.deviceKeyInto(fk, folder, vl.Versions[0].Device, name)
+
+		f, ok := db.getFileTrunc(fk, truncate)
+		if !ok {
 			continue
 		}
 
-		if cont := fn(f); !cont {
+		if !fn(f) {
 			return
 		}
 	}
@@ -316,10 +384,8 @@ func (db *Instance) availability(folder, file []byte) []protocol.DeviceID {
 		return nil
 	}
 
-	var vl VersionList
-	err = vl.Unmarshal(bs)
-	if err != nil {
-		l.Debugln("unmarshal error:", err)
+	vl, ok := unmarshalVersionList(bs)
+	if !ok {
 		return nil
 	}
 
@@ -339,6 +405,11 @@ func (db *Instance) availability(folder, file []byte) []protocol.DeviceID {
 }
 
 func (db *Instance) withNeed(folder, device []byte, truncate bool, fn Iterator) {
+	if bytes.Equal(device, protocol.LocalDeviceID[:]) {
+		db.withNeedLocal(folder, truncate, fn)
+		return
+	}
+
 	t := db.newReadOnlyTransaction()
 	defer t.close()
 
@@ -347,33 +418,16 @@ func (db *Instance) withNeed(folder, device []byte, truncate bool, fn Iterator)
 
 	var fk []byte
 	for dbi.Next() {
-		var vl VersionList
-		err := vl.Unmarshal(dbi.Value())
-		if err != nil {
-			l.Debugln("unmarshal error:", err)
-			continue
-		}
-		if len(vl.Versions) == 0 {
-			l.Debugln("no versions:", dbi.Key())
+		vl, ok := unmarshalVersionList(dbi.Value())
+		if !ok {
 			continue
 		}
 
-		have := false // If we have the file, any version
-		need := false // If we have a lower version of the file
-		var haveFileVersion FileVersion
-		for _, v := range vl.Versions {
-			if bytes.Equal(v.Device, device) {
-				have = true
-				haveFileVersion = v
-				// XXX: This marks Concurrent (i.e. conflicting) changes as
-				// needs. Maybe we should do that, but it needs special
-				// handling in the puller.
-				need = !v.Version.GreaterEqual(vl.Versions[0].Version)
-				break
-			}
-		}
-
-		if have && !need {
+		haveFV, have := vl.Get(device)
+		// XXX: This marks Concurrent (i.e. conflicting) changes as
+		// needs. Maybe we should do that, but it needs special
+		// handling in the puller.
+		if have && haveFV.Version.GreaterEqual(vl.Versions[0].Version) {
 			continue
 		}
 
@@ -392,7 +446,7 @@ func (db *Instance) withNeed(folder, device []byte, truncate bool, fn Iterator)
 				continue
 			}
 
-			fk = db.deviceKeyInto(fk[:cap(fk)], folder, vl.Versions[i].Device, name)
+			fk = db.deviceKeyInto(fk, folder, vl.Versions[i].Device, name)
 			bs, err := t.Get(fk, nil)
 			if err != nil {
 				l.Debugln("surprise error:", err)
@@ -410,9 +464,9 @@ func (db *Instance) withNeed(folder, device []byte, truncate bool, fn Iterator)
 				break
 			}
 
-			l.Debugf("need folder=%q device=%v name=%q need=%v have=%v invalid=%v haveV=%v globalV=%v globalDev=%v", folder, protocol.DeviceIDFromBytes(device), name, need, have, haveFileVersion.Invalid, haveFileVersion.Version, needVersion, needDevice)
+			l.Debugf("need folder=%q device=%v name=%q have=%v invalid=%v haveV=%v globalV=%v globalDev=%v", folder, protocol.DeviceIDFromBytes(device), name, have, haveFV.Invalid, haveFV.Version, needVersion, needDevice)
 
-			if cont := fn(gf); !cont {
+			if !fn(gf) {
 				return
 			}
 
@@ -422,6 +476,28 @@ func (db *Instance) withNeed(folder, device []byte, truncate bool, fn Iterator)
 	}
 }
 
+func (db *Instance) withNeedLocal(folder []byte, truncate bool, fn Iterator) {
+	t := db.newReadOnlyTransaction()
+	defer t.close()
+
+	dbi := t.NewIterator(util.BytesPrefix(db.needKey(folder, nil)[:keyPrefixLen+keyFolderLen]), nil)
+	defer dbi.Release()
+
+	var dk []byte
+	var gk []byte
+	var f FileIntf
+	var ok bool
+	for dbi.Next() {
+		gk, dk, f, ok = db.getGlobalInto(t, gk, dk, folder, db.globalKeyName(dbi.Key()), truncate)
+		if !ok {
+			continue
+		}
+		if !fn(f) {
+			return
+		}
+	}
+}
+
 func (db *Instance) ListFolders() []string {
 	t := db.newReadOnlyTransaction()
 	defer t.close()
@@ -447,28 +523,21 @@ func (db *Instance) ListFolders() []string {
 }
 
 func (db *Instance) dropFolder(folder []byte) {
-	t := db.newReadOnlyTransaction()
+	t := db.newReadWriteTransaction()
 	defer t.close()
 
-	// Remove all items related to the given folder from the device->file bucket
-	dbi := t.NewIterator(util.BytesPrefix([]byte{KeyTypeDevice}), nil)
-	for dbi.Next() {
-		itemFolder := db.deviceKeyFolder(dbi.Key())
-		if bytes.Equal(folder, itemFolder) {
-			db.Delete(dbi.Key(), nil)
-		}
+	for _, key := range [][]byte{
+		// Remove all items related to the given folder from the device->file bucket
+		db.deviceKey(folder, nil, nil)[:keyPrefixLen+keyFolderLen],
+		// Remove all sequences related to the folder
+		db.sequenceKey([]byte(folder), 0)[:keyPrefixLen+keyFolderLen],
+		// Remove all items related to the given folder from the global bucket
+		db.globalKey(folder, nil)[:keyPrefixLen+keyFolderLen],
+		// Remove all needs related to the folder
+		db.needKey(folder, nil)[:keyPrefixLen+keyFolderLen],
+	} {
+		t.deleteKeyPrefix(key)
 	}
-	dbi.Release()
-
-	// Remove all items related to the given folder from the global bucket
-	dbi = t.NewIterator(util.BytesPrefix([]byte{KeyTypeGlobal}), nil)
-	for dbi.Next() {
-		itemFolder, ok := db.globalKeyFolder(dbi.Key())
-		if ok && bytes.Equal(folder, itemFolder) {
-			db.Delete(dbi.Key(), nil)
-		}
-	}
-	dbi.Release()
 }
 
 func (db *Instance) dropDeviceFolder(device, folder []byte, meta *metadataTracker) {
@@ -478,10 +547,13 @@ func (db *Instance) dropDeviceFolder(device, folder []byte, meta *metadataTracke
 	dbi := t.NewIterator(util.BytesPrefix(db.deviceKey(folder, device, nil)), nil)
 	defer dbi.Release()
 
+	var gk []byte
+
 	for dbi.Next() {
 		key := dbi.Key()
 		name := db.deviceKeyName(key)
-		t.removeFromGlobal(folder, device, name, meta)
+		gk = db.globalKeyInto(gk, folder, name)
+		t.removeFromGlobal(gk, folder, device, name, meta)
 		t.Delete(key)
 		t.checkFlush()
 	}
@@ -496,11 +568,8 @@ func (db *Instance) checkGlobals(folder []byte, meta *metadataTracker) {
 
 	var fk []byte
 	for dbi.Next() {
-		gk := dbi.Key()
-		var vl VersionList
-		err := vl.Unmarshal(dbi.Value())
-		if err != nil {
-			l.Debugln("unmarshal error:", err)
+		vl, ok := unmarshalVersionList(dbi.Value())
+		if !ok {
 			continue
 		}
 
@@ -509,11 +578,10 @@ func (db *Instance) checkGlobals(folder []byte, meta *metadataTracker) {
 		// there are global entries pointing to no longer existing files. Here
 		// we find those and clear them out.
 
-		name := db.globalKeyName(gk)
+		name := db.globalKeyName(dbi.Key())
 		var newVL VersionList
 		for i, version := range vl.Versions {
-			fk = db.deviceKeyInto(fk[:cap(fk)], folder, version.Device, name)
-
+			fk = db.deviceKeyInto(fk, folder, version.Device, name)
 			_, err := t.Get(fk, nil)
 			if err == leveldb.ErrNotFound {
 				continue
@@ -525,8 +593,8 @@ func (db *Instance) checkGlobals(folder []byte, meta *metadataTracker) {
 			newVL.Versions = append(newVL.Versions, version)
 
 			if i == 0 {
-				if fi, ok := t.getFile(folder, version.Device, name); ok {
-					meta.addFile(globalDeviceID, fi)
+				if fi, ok := db.getFile(fk); ok {
+					meta.addFile(protocol.GlobalDeviceID, fi)
 				}
 			}
 		}
@@ -539,73 +607,6 @@ func (db *Instance) checkGlobals(folder []byte, meta *metadataTracker) {
 	l.Debugf("db check completed for %q", folder)
 }
 
-func (db *Instance) updateSchema0to1() {
-	t := db.newReadWriteTransaction()
-	defer t.close()
-
-	dbi := t.NewIterator(util.BytesPrefix([]byte{KeyTypeDevice}), nil)
-	defer dbi.Release()
-
-	symlinkConv := 0
-	changedFolders := make(map[string]struct{})
-	ignAdded := 0
-	meta := newMetadataTracker() // dummy metadata tracker
-
-	for dbi.Next() {
-		folder := db.deviceKeyFolder(dbi.Key())
-		device := db.deviceKeyDevice(dbi.Key())
-		name := string(db.deviceKeyName(dbi.Key()))
-
-		// Remove files with absolute path (see #4799)
-		if strings.HasPrefix(name, "/") {
-			if _, ok := changedFolders[string(folder)]; !ok {
-				changedFolders[string(folder)] = struct{}{}
-			}
-			t.removeFromGlobal(folder, device, nil, nil)
-			t.Delete(dbi.Key())
-			t.checkFlush()
-			continue
-		}
-
-		// Change SYMLINK_FILE and SYMLINK_DIRECTORY types to the current SYMLINK
-		// type (previously SYMLINK_UNKNOWN). It does this for all devices, both
-		// local and remote, and does not reset delta indexes. It shouldn't really
-		// matter what the symlink type is, but this cleans it up for a possible
-		// future when SYMLINK_FILE and SYMLINK_DIRECTORY are no longer understood.
-		var f protocol.FileInfo
-		if err := f.Unmarshal(dbi.Value()); err != nil {
-			// probably can't happen
-			continue
-		}
-		if f.Type == protocol.FileInfoTypeDeprecatedSymlinkDirectory || f.Type == protocol.FileInfoTypeDeprecatedSymlinkFile {
-			f.Type = protocol.FileInfoTypeSymlink
-			bs, err := f.Marshal()
-			if err != nil {
-				panic("can't happen: " + err.Error())
-			}
-			t.Put(dbi.Key(), bs)
-			t.checkFlush()
-			symlinkConv++
-		}
-
-		// Add invalid files to global list
-		if f.Invalid {
-			if t.updateGlobal(folder, device, f, meta) {
-				if _, ok := changedFolders[string(folder)]; !ok {
-					changedFolders[string(folder)] = struct{}{}
-				}
-				ignAdded++
-			}
-		}
-	}
-
-	for folder := range changedFolders {
-		db.dropFolderMeta([]byte(folder))
-	}
-
-	l.Infof("Updated symlink type for %d index entries and added %d invalid files to global list", symlinkConv, ignAdded)
-}
-
 // deviceKey returns a byte slice encoding the following information:
 //	   keyTypeDevice (1 byte)
 //	   folder (4 bytes)
@@ -615,16 +616,14 @@ func (db *Instance) deviceKey(folder, device, file []byte) []byte {
 	return db.deviceKeyInto(nil, folder, device, file)
 }
 
-func (db *Instance) deviceKeyInto(k []byte, folder, device, file []byte) []byte {
+func (db *Instance) deviceKeyInto(k, folder, device, file []byte) []byte {
 	reqLen := keyPrefixLen + keyFolderLen + keyDeviceLen + len(file)
-	if len(k) < reqLen {
-		k = make([]byte, reqLen)
-	}
+	k = resize(k, reqLen)
 	k[0] = KeyTypeDevice
 	binary.BigEndian.PutUint32(k[keyPrefixLen:], db.folderIdx.ID(folder))
 	binary.BigEndian.PutUint32(k[keyPrefixLen+keyFolderLen:], db.deviceIdx.ID(device))
 	copy(k[keyPrefixLen+keyFolderLen+keyDeviceLen:], file)
-	return k[:reqLen]
+	return k
 }
 
 // deviceKeyName returns the device ID from the key
@@ -655,11 +654,16 @@ func (db *Instance) deviceKeyDevice(key []byte) []byte {
 //	   folder (4 bytes)
 //	   name (variable size)
 func (db *Instance) globalKey(folder, file []byte) []byte {
-	k := make([]byte, keyPrefixLen+keyFolderLen+len(file))
-	k[0] = KeyTypeGlobal
-	binary.BigEndian.PutUint32(k[keyPrefixLen:], db.folderIdx.ID(folder))
-	copy(k[keyPrefixLen+keyFolderLen:], file)
-	return k
+	return db.globalKeyInto(nil, folder, file)
+}
+
+func (db *Instance) globalKeyInto(gk, folder, file []byte) []byte {
+	reqLen := keyPrefixLen + keyFolderLen + len(file)
+	gk = resize(gk, reqLen)
+	gk[0] = KeyTypeGlobal
+	binary.BigEndian.PutUint32(gk[keyPrefixLen:], db.folderIdx.ID(folder))
+	copy(gk[keyPrefixLen+keyFolderLen:], file)
+	return gk
 }
 
 // globalKeyName returns the filename from the key
@@ -672,6 +676,39 @@ func (db *Instance) globalKeyFolder(key []byte) ([]byte, bool) {
 	return db.folderIdx.Val(binary.BigEndian.Uint32(key[keyPrefixLen:]))
 }
 
+// needKey is a globalKey with a different prefix
+func (db *Instance) needKey(folder, file []byte) []byte {
+	return db.needKeyInto(nil, folder, file)
+}
+
+func (db *Instance) needKeyInto(k, folder, file []byte) []byte {
+	k = db.globalKeyInto(k, folder, file)
+	k[0] = KeyTypeNeed
+	return k
+}
+
+// sequenceKey returns a byte slice encoding the following information:
+//	   KeyTypeSequence (1 byte)
+//	   folder (4 bytes)
+//	   sequence number (8 bytes)
+func (db *Instance) sequenceKey(folder []byte, seq int64) []byte {
+	return db.sequenceKeyInto(nil, folder, seq)
+}
+
+func (db *Instance) sequenceKeyInto(k []byte, folder []byte, seq int64) []byte {
+	reqLen := keyPrefixLen + keyFolderLen + keySequenceLen
+	k = resize(k, reqLen)
+	k[0] = KeyTypeSequence
+	binary.BigEndian.PutUint32(k[keyPrefixLen:], db.folderIdx.ID(folder))
+	binary.BigEndian.PutUint64(k[keyPrefixLen+keyFolderLen:], uint64(seq))
+	return k
+}
+
+// sequenceKeySequence returns the sequence number from the key
+func (db *Instance) sequenceKeySequence(key []byte) int64 {
+	return int64(binary.BigEndian.Uint64(key[keyPrefixLen+keyFolderLen:]))
+}
+
 func (db *Instance) getIndexID(device, folder []byte) protocol.IndexID {
 	key := db.indexIDKey(device, folder)
 	cur, err := db.Get(key, nil)
@@ -787,6 +824,19 @@ func unmarshalTrunc(bs []byte, truncate bool) (FileIntf, error) {
 	return tf, err
 }
 
+func unmarshalVersionList(data []byte) (VersionList, bool) {
+	var vl VersionList
+	if err := vl.Unmarshal(data); err != nil {
+		l.Debugln("unmarshal error:", err)
+		return VersionList{}, false
+	}
+	if len(vl.Versions) == 0 {
+		l.Debugln("empty version list")
+		return VersionList{}, false
+	}
+	return vl, true
+}
+
 // A "better" version of leveldb's errors.IsCorrupted.
 func leveldbIsCorrupted(err error) bool {
 	switch {
@@ -887,3 +937,20 @@ func (i *smallIndex) Val(id uint32) ([]byte, bool) {
 
 	return []byte(val), true
 }
+
+// resize returns a byte array of length reqLen, reusing k if possible
+func resize(k []byte, reqLen int) []byte {
+	if cap(k) < reqLen {
+		return make([]byte, reqLen)
+	}
+	return k[:reqLen]
+}
+
+type errorSuggestion struct {
+	inner      error
+	suggestion string
+}
+
+func (e errorSuggestion) Error() string {
+	return fmt.Sprintf("%s (%s)", e.inner.Error(), e.suggestion)
+}
